﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HEAL.Attic;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Regression {
  [Item("Pearson R² Constraint Evaluator", "Calculates the square of the pearson correlation coefficient (also known as coefficient of determination) of a symbolic regression solution.")]
  [StorableType("D61462E4-2032-4790-B63D-5E6512987F64")]
  public class SymbolicRegressionSingleObjectiveConstraintPearsonRSquaredEvaluator : SymbolicRegressionSingleObjectiveEvaluator {
    [StorableConstructor]
    protected SymbolicRegressionSingleObjectiveConstraintPearsonRSquaredEvaluator(StorableConstructorFlag _) : base(_) { }
    protected SymbolicRegressionSingleObjectiveConstraintPearsonRSquaredEvaluator(SymbolicRegressionSingleObjectiveConstraintPearsonRSquaredEvaluator original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicRegressionSingleObjectiveConstraintPearsonRSquaredEvaluator(this, cloner);
    }

    public SymbolicRegressionSingleObjectiveConstraintPearsonRSquaredEvaluator() : base() { }

    public override bool Maximization { get { return true; } }

    public override IOperation InstrumentedApply() {
      var solution = SymbolicExpressionTreeParameter.ActualValue;
      IEnumerable<int> rows = GenerateRowsToEvaluate();

      double quality = Calculate(SymbolicDataAnalysisTreeInterpreterParameter.ActualValue, solution, EstimationLimitsParameter.ActualValue.Lower, EstimationLimitsParameter.ActualValue.Upper, ProblemDataParameter.ActualValue, rows, ApplyLinearScalingParameter.ActualValue.Value);
      QualityParameter.ActualValue = new DoubleValue(quality);
      return base.InstrumentedApply();
    }

    public static double Calculate(ISymbolicDataAnalysisExpressionTreeInterpreter interpreter, ISymbolicExpressionTree solution, double lowerEstimationLimit, double upperEstimationLimit, IRegressionProblemData problemData, IEnumerable<int> rows, bool applyLinearScaling) {
      IEnumerable<double> estimatedValues = interpreter.GetSymbolicExpressionTreeValues(solution, problemData.Dataset, rows);
      IEnumerable<double> targetValues = problemData.Dataset.GetDoubleValues(problemData.TargetVariable, rows);
      OnlineCalculatorError errorState = OnlineCalculatorError.None;

      IntervalConstraintsParser parser = new IntervalConstraintsParser();
      var constraints = parser.Parse(((RegressionProblemData) problemData).IntervalConstraintsParameter.Value.Value);
      var intervalInterpreter = new IntervalInterpreter();
      var variableRanges = ((RegressionProblemData) problemData).VariableRangesParameter.Value.VariableIntervals;


      //foreach (var constraint in constraints) {
      //  if (constraint.Variable != null && !variableRanges.ContainsKey(constraint.Variable))
      //    throw new ArgumentException($"The given variable {constraint.Variable} in the constraint does not exists in the model.", nameof(IntervalConstraintsParser));
      //  if (!constraint.IsDerivation) {
      //    var res = intervalInterpreter.GetSymbolicExressionTreeInterval(solution, variableRanges);
      //    if (!constraint.Interval.Contains(res, constraint.InclusiveLowerBound,
      //      constraint.InclusiveUpperBound)) {
      //      return 0;
      //    }
      //  } else {
      //    var tree = solution;
      //    for (var i = 0; i < constraint.NumberOfDerivation; ++i) {
      //      tree = DerivativeCalculator.Derive(tree, constraint.Variable);
      //    }
      //    var res = intervalInterpreter.GetSymbolicExressionTreeInterval(tree, variableRanges);
      //    if (!constraint.Interval.Contains(res, constraint.InclusiveLowerBound,
      //      constraint.InclusiveUpperBound)) {
      //      return 0;
      //    }
      //  }
      //}
      // TODO
      // m = new SymbolicRegressionModel(...)
      // m.Scale();

      // var e = m.GetEstimatedValues (TRAINING)
      // OnlinePearsonCalc.Calculate(e, TARGET_TRAIING)
      
      // scaledTree = model.Tree;

      // constraints mit scaledTree berechnen (auch die Ableitungen)

      double r;
      if (applyLinearScaling) {
        var rCalculator = new OnlinePearsonsRCalculator();
        CalculateWithScaling(targetValues, estimatedValues, lowerEstimationLimit, upperEstimationLimit, rCalculator, problemData.Dataset.Rows);
        var model = new SymbolicRegressionModel(problemData.TargetVariable, solution, interpreter, lowerEstimationLimit, upperEstimationLimit);
        model.Scale(problemData);
        var e = model.GetEstimatedValues(problemData.Dataset, problemData.TrainingIndices);
        var scaledTree = model.SymbolicExpressionTree;
        errorState = rCalculator.ErrorState;
        if (CheckConstraintsViolations(constraints, intervalInterpreter, variableRanges, scaledTree, out var val)) {
          r = val;
        } else {
          r = rCalculator.R;
        }
      } else {
        IEnumerable<double> boundedEstimatedValues = estimatedValues.LimitToRange(lowerEstimationLimit, upperEstimationLimit);
        if (CheckConstraintsViolations(constraints, intervalInterpreter, variableRanges, solution, out var val)) {
          r = val;
        } else {
          r = OnlinePearsonsRCalculator.Calculate(targetValues, boundedEstimatedValues, out errorState);
        }
      }
      if (errorState != OnlineCalculatorError.None) return double.NaN;
      return r*r;
    }


    public override double Evaluate(IExecutionContext context, ISymbolicExpressionTree tree,
      IRegressionProblemData problemData, IEnumerable<int> rows) {
      SymbolicDataAnalysisTreeInterpreterParameter.ExecutionContext = context;
      EstimationLimitsParameter.ExecutionContext = context;
      ApplyLinearScalingParameter.ExecutionContext = context;

      double r2 = Calculate(SymbolicDataAnalysisTreeInterpreterParameter.ActualValue, tree,
        EstimationLimitsParameter.ActualValue.Lower, EstimationLimitsParameter.ActualValue.Upper, problemData, rows,
        ApplyLinearScalingParameter.ActualValue.Value);

      SymbolicDataAnalysisTreeInterpreterParameter.ExecutionContext = null;
      EstimationLimitsParameter.ExecutionContext = null;
      ApplyLinearScalingParameter.ExecutionContext = null;

      return r2;
    }

    private static bool CheckConstraintsViolations(List<IntervalConstraint> constraints, IntervalInterpreter intervalInterpreter,
      Dictionary<string, Interval> variableRanges, ISymbolicExpressionTree solution, out double r) {
      foreach (var constraint in constraints) {
        if (constraint.Variable != null && !variableRanges.ContainsKey(constraint.Variable))
          throw new ArgumentException($"The given variable {constraint.Variable} in the constraint does not exists in the model.", nameof(IntervalConstraintsParser));
        if (!constraint.IsDerivation) {
          var res = intervalInterpreter.GetSymbolicExressionTreeInterval(solution, variableRanges);
          if (!constraint.Interval.Contains(res, constraint.InclusiveLowerBound,
            constraint.InclusiveUpperBound)) {
            r = 0;
            return true;
          }
        } else {
          var tree = solution;
          for (var i = 0; i < constraint.NumberOfDerivation; ++i) {
            tree = DerivativeCalculator.Derive(tree, constraint.Variable);
          }
          var res = intervalInterpreter.GetSymbolicExressionTreeInterval(tree, variableRanges);
          if (!constraint.Interval.Contains(res, constraint.InclusiveLowerBound,
            constraint.InclusiveUpperBound)) {
            r = 0;
            return true;
          }
        }
      }
      r = 1;
      return false;
    }
  }
}
